/*
 * Title:        sag_internal/hashes.hpp
 * Description:  Implementation of hash on our data types
 * $Copyright (c) 2015-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: hashes.hpp 334273 2018-07-24 07:32:42Z tdi $
 */


#ifndef _SAG_INTERNAL_HASHES_HPP_
#define _SAG_INTERNAL_HASHES_HPP_

#include <sag_internal/murmur_64a.hpp>

// already in com.softwareag.connectivity

namespace hashes {

typedef uint64_t hash_t;

hash_t hash(const list_t &list);
hash_t hash(const map_t &map);
hash_t hash(const buffer_t &buffer);
hash_t hash(const char *str);

struct calculate_hash
{
	static uint64_t SEED() { return 0x243FCAF1; } // just a random number
	uint64_t seed;
	calculate_hash(uint64_t seed=SEED()): seed(seed) {}
	struct result_t { int64_t h1; int64_t h2; };
	typedef hash_t result_type;
	hash_t operator()() const {
		return seed;
//		throw std::runtime_error("Cannot use null as a key");
	}
	hash_t operator()(int64_t i) const {
		return murmur_hash_process64a(reinterpret_cast<const char *>(&i),sizeof(i), seed) ;  
	}
	hash_t operator()(decimal_t i) const {
		return murmur_hash_process64a(reinterpret_cast<const char *>(&i),sizeof(i), seed) ;
	}
	hash_t operator()(double d) const {
		// map -0.0 and 0.0 to 0 so give the same hash.
		if (d==-0.0) {
			d = 0.0 ;
		}
		return murmur_hash_process64a(reinterpret_cast<const char *>(&d),sizeof(d), seed) ;	  
	}
	hash_t operator()(bool b) const {
		return b ? seed : ~seed;
	}
	hash_t operator()(const char *s) const {
		return hash(s);
	}
	hash_t operator()(const sag_underlying_custom_t &) const {
		throw std::runtime_error("Cannot use custom values as a key");
	}
	hash_t operator()(const map_t &m) const {
		return hash(m);
	}
	hash_t operator()(const list_t &m) const {
		return hash(m);
	}
	hash_t operator()(const buffer_t &m) const {
		return hash(m);
	}

	hash_t operator()(const data_t &d) const {
		return apply_visitor(*this, d);
	}

};

inline hash_t hash(const char *str)
{
	return sag_hash_string(str);
}
inline hash_t hash(const data_t &data, uint64_t seed=calculate_hash::SEED())
{
	return apply_visitor(calculate_hash(seed), data);
}

inline hash_t hash(const list_t &list) 
{
	hash_t h = calculate_hash::SEED();
	for (list_t::const_iterator it = list.begin(); it != list.end(); ++it) {
		h = hash(*it, (uint64_t) h);
	}
	return h;
} 

inline hash_t hash(const map_t &map)
{
	hash_t h = calculate_hash::SEED();
	for (map_t::const_iterator it = map.begin(); it != map.end(); ++it) {
		h += hash(it->first);
		h += hash(it->second);
	}
	return h;
} 

inline hash_t hash(const buffer_t &buffer) 
{  
	return murmur_hash_process64a(reinterpret_cast<const char *>(buffer.begin()), buffer.size(), calculate_hash::SEED()) ;
}

} // hashes

#endif // _SAG_INTERNAL_HASHES_HPP_
