/*
 * Title:        sag_internal/murmur_64a.hpp
 * Description:  Implementation of murmur_64a algorithm for murmur_hash2 version 1.1.1.
 * @Version:     $Id: murmur_64a.hpp 335831 2018-08-15 13:47:36Z tdi $
 */


#ifndef _MURMURHASH_64A_H_
#define _MURMURHASH_64A_H_


// already in namespace com::softwareag::connectivity



/****************************************************************
 * Copyright (c) 2013 ksss
 * MIT License

 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
 * documentation files (the"Software"), to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
 * and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all copies or substantial portions 
 * of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 *
 * MurmurHash64A (C) Austin Appleby
 */

#if defined(_MSC_VER)

#define BIG_CONSTANT_A64(x) (x)

// Other compilers

#else	// defined(_MSC_VER)

#define BIG_CONSTANT_A64(x) (x##LLU)

#endif // !defined(_MSC_VER)


 
inline uint64_t murmur_hash_process64a(const char *key, size_t len, uint64_t seed)     // hashing optimized to 64bit hashing on 64bit systems.
{

  const uint64_t m = BIG_CONSTANT_A64(0xc6a4a7935bd1e995);
  const int r = 47;

  uint64_t h = seed ^ (len * m);

  const uint64_t * data = (const uint64_t *)key;
  const uint64_t * end = data + (len/8);
  
  while(data != end)
  {
    uint64_t k = *data++;
    
    k *= m;
    k ^= k >> r;
    k *= m;

    h ^= k;
    h *= m;
  }

  const unsigned char * data2 = (const unsigned char*)data;

  switch(len & 7) {
    case 7: h ^= (uint64_t)((uint64_t)data2[6] << (uint64_t)48);
    case 6: h ^= (uint64_t)((uint64_t)data2[5] << (uint64_t)40);
    case 5: h ^= (uint64_t)((uint64_t)data2[4] << (uint64_t)32);
    case 4: h ^= (uint64_t)((uint64_t)data2[3] << (uint64_t)24);
    case 3: h ^= (uint64_t)((uint64_t)data2[2] << (uint64_t)16);
    case 2: h ^= (uint64_t)((uint64_t)data2[1] << (uint64_t)8 );
    case 1: h ^= (uint64_t)((uint64_t)data2[0]                );
            h *= m;
  };

  h ^= h >> r;
  h *= m;
  h ^= h >> r;

  return h;
}


#endif // _MURMURHASH_A64_H_

